import asyncio
import logging
from aiohttp.web import FileField
from tukio import get_broker, EXEC_TOPIC
from tukio.utils import FutureState
from tukio.workflow import (
    WorkflowTemplate, WorkflowExecState, WorkflowRootTaskError
)
from pymongo.errors import AutoReconnect

from nyuki.api import Response, resource, content_type, HTTPBreak
from nyuki.utils import from_isoformat
from nyuki.workflow.tasks.utils.uri import URI, InvalidWorkflowUri
from nyuki.workflow.db.workflow_instances import Ordering


log = logging.getLogger(__name__)


class _WorkflowResource:

    """
    Share methods between workflow resources
    """

    def register_async_handler(self, async_topic, events, wflow):
        broker = get_broker()
        topic = '/'.join((EXEC_TOPIC, wflow.uid))

        if events:
            events = events.split(',')

        async def exec_handler(event):
            # Publish the event's data if requested.
            if not events or event.data['type'] in events:
                await self.nyuki.bus.publish(event.data, async_topic)
            # If the workflow is in a final state, unregister
            if event.data['type'] in [
                WorkflowExecState.END.value,
                WorkflowExecState.ERROR.value
            ]:
                broker.unregister(exec_handler, topic=topic)

        broker.register(exec_handler, topic=topic)


@resource('/workflow/instances', ['v1'], 'application/json')
class ApiWorkflows(_WorkflowResource):

    async def get(self, request):
        """
        Return workflow instances
        """
        workflows = []
        children = request.query.get('children', '0') == '1'
        tasks = request.query.get('tasks', '0') == '1'

        for wflow in self.nyuki.running_workflows.values():
            if children is False:
                requester = wflow.exec.get('requester')
                if requester and requester.startswith('nyuki://'):
                    continue
            workflows.append(wflow.report(tasks=tasks))

        return Response(workflows)

    async def put(self, request):
        """
        Start a workflow from payload:
        {
            "id": "template_id",
            "draft": true/false,
            "exec": {}
        }
        """
        async_topic = request.headers.get('X-Surycat-Async-Topic')
        async_events = request.headers.get('X-Surycat-Async-Events')
        exec_track = request.headers.get('X-Surycat-Exec-Track')
        requester = request.headers.get('Referer')
        request = await request.json()

        if 'id' not in request:
            return Response(status=400, body={
                'error': "Template's ID key 'id' is mandatory"
            })
        draft = request.get('draft', False)
        data = request.get('inputs', {})
        exec = request.get('exec')

        if exec:
            # Suspended/crashed instance
            # The request's payload is the last known execution report
            if exec['id'] in self.nyuki.running_workflows:
                return Response(status=400, body={
                    'error': 'This workflow is already being rescued'
                })
        else:
            # Fetch the template from the storage
            try:
                template = await self.nyuki.storage.get_template(
                    request['id'], draft=draft
                )
            except AutoReconnect:
                return Response(status=503)

        if not template:
            return Response(status=404, body={
                'error': 'Could not find a suitable template to run'
            })

        wf_tmpl = WorkflowTemplate.from_dict(template)
        try:
            wf_tmpl.root()
        except WorkflowRootTaskError:
            return Response(status=400, body={
                'error': 'More than one root task'
            })

        if exec:
            wflow = await self.nyuki.engine.rescue(wf_tmpl, request)
        elif draft:
            wflow = await self.nyuki.engine.run_once(wf_tmpl, data)
        else:
            wflow = await self.nyuki.engine.trigger(wf_tmpl.uid, data)

        if wflow is None:
            return Response(status=400, body={
                'error': 'Could not start any workflow from this template'
            })

        # Prevent workflow loop
        exec_track = exec_track.split(',') if exec_track else []
        holder = self.nyuki.bus.name
        for ancestor in exec_track:
            try:
                info = URI.parse(ancestor)
            except InvalidWorkflowUri:
                continue
            if info.template_id == wf_tmpl.uid and info.holder == holder:
                return Response(status=400, body={
                    'error': 'Loop detected between workflows'
                })

        # Keep full instance+template in nyuki's memory
        wfinst = self.nyuki.new_workflow(
            template, wflow,
            track=exec_track,
            requester=requester
        )
        # Handle async workflow exec updates
        if async_topic is not None:
            self.register_async_handler(async_topic, async_events, wflow)

        try:
            # Wait up to 30 seconds for the workflow to start.
            await asyncio.wait_for(wfinst.instance._committed.wait(), 30.0)
        except asyncio.TimeoutError:
            status = 201
        else:
            status = 200

        return Response(wfinst.report(), status=status)


@resource('/workflow/instances/{iid}', versions=['v1'])
class ApiWorkflow:

    async def get(self, request, iid):
        """
        Return a workflow instance
        """
        try:
            return Response(self.nyuki.running_workflows[iid].report(data=False))
        except KeyError:
            return Response(status=404)

    async def post(self, request, iid):
        """
        Suspend/resume a runnning workflow.
        """
        try:
            wf = self.nyuki.running_workflows[iid]
        except KeyError:
            return Response(status=404)

        request = await request.json()

        try:
            action = request['action']
        except KeyError:
            return Response(status=400, body={
                'action parameter required'
            })

        # Should we return 409 Conflict if the status is already set ?
        if action == 'suspend':
            wf.instance.suspend()
        elif action == 'resume':
            wf.instance.resume()
        else:
            return Response(status=400, body={
                "action must be 'suspend' or 'resume'"
            })

        return Response(wf.report())

    async def delete(self, request, iid):
        """
        Cancel a workflow instance.
        """
        try:
            self.nyuki.running_workflows[iid].instance.cancel()
        except KeyError:
            return Response(status=404)


@resource('/workflow/instances/{iid}/tasks/{tid}/reporting', versions=['v1'])
class ApiTaskReporting:

    async def get(self, request, iid, tid):
        """
        Return all the current reporting data for this task.
        """
        try:
            workflow = self.nyuki.running_workflows[iid].instance
            for task in workflow.tasks:
                if task.uid == tid:
                    return Response(task.holder.report() or {})
            else:
                return Response(status=404)
        except KeyError:
            return Response(status=404)


@resource('/workflow/instances/{iid}/tasks/{tid}/reporting/contacts', versions=['v1'])
class ApiTaskReportingContacts:

    async def get(self, request, iid, tid):
        """
        Return all contacts' informations from the reporting data of this task.
        """
        try:
            workflow = self.nyuki.running_workflows[iid].instance
            for task in workflow.tasks:
                if task.uid == tid:
                    if not hasattr(task.holder, 'report_contacts'):
                        raise HTTPBreak(404)
                    return Response(list(task.holder.report_contacts().values()))
            else:
                raise HTTPBreak(404)
        except KeyError:
            raise HTTPBreak(404)


@resource('/workflow/instances/{iid}/tasks/{tid}/reporting/contacts/{cuid}', versions=['v1'])
class ApiTaskReportingContact:

    async def get(self, request, iid, tid, cuid):
        """
        Return one contact's informations from the reporting data of this task.
        """
        try:
            workflow = self.nyuki.running_workflows[iid].instance
            for task in workflow.tasks:
                if task.uid == tid:
                    if not hasattr(task.holder, 'report_contacts'):
                        raise HTTPBreak(404)
                    contacts = task.holder.report_contacts()
                    return Response(contacts[cuid])
            else:
                raise HTTPBreak(404)
        except KeyError:
            raise HTTPBreak(404)


@resource('/workflow/history', versions=['v1'])
class ApiWorkflowsHistory:

    async def get(self, request):
        """
        Filters:
            * `root` return only the root workflows
            * `full` return the full graph and details of all workflows
                * :warning: can be a huge amount of data
            * `since` return the workflows since this date
            * `state` return the workflows on this FutureState
            * `offset` return the worflows from this offset
            * `limit` return this amount of workflows
            * `order` order results following the Ordering enum values
            * `search` search templates with specific title
        """
        # Filter on start date
        since = request.query.get('since')
        if since:
            try:
                since = from_isoformat(since)
            except ValueError:
                return Response(status=400, body={
                    'error': "Could not parse date '{}'".format(since)
                })
        # Filter on state value
        state = request.query.get('state')
        if state:
            try:
                state = FutureState(state)
            except ValueError:
                return Response(status=400, body={
                    'error': "Unknown state '{}'".format(state)
                })
        # Skip first items
        offset = request.query.get('offset')
        if offset:
            try:
                offset = int(offset)
            except ValueError:
                return Response(status=400, body={
                    'error': 'Offset must be an int'
                })
        # Limit max result
        limit = request.query.get('limit')
        if limit:
            try:
                limit = int(limit)
            except ValueError:
                return Response(status=400, body={
                    'error': 'Limit must be an int'
                })
        order = request.query.get('ordering')
        if order:
            try:
                order = Ordering[order].value
            except KeyError:
                return Response(status=400, body={
                    'error': 'Ordering must be in {}'.format(Ordering.keys())
                })

        try:
            count, history = await self.nyuki.storage.get_history(
                root=(request.query.get('root') == '1'),
                full=(request.query.get('full') == '1'),
                search=request.query.get('search'),
                order=order,
                offset=offset, limit=limit, since=since, state=state,
            )
        except AutoReconnect:
            return Response(status=503)

        data = {'count': count, 'data': history}
        return Response(data)


@resource('/workflow/history/{uid}', versions=['v1'])
class ApiWorkflowHistory:

    async def get(self, request, uid):
        try:
            workflow = await self.nyuki.storage.get_instance(
                uid, (request.query.get('full') == '1')
            )
        except AutoReconnect:
            return Response(status=503)
        if not workflow:
            return Response(status=404)
        return Response(workflow)


@resource('/workflow/history/{uid}/tasks/{task_id}', versions=['v1'])
class ApiWorkflowHistoryTask:

    async def get(self, request, uid, task_id):
        try:
            task = await self.nyuki.storage.get_instance_task(
                task_id, (request.query.get('full') == '1')
            )
        except AutoReconnect:
            return Response(status=503)
        if not task:
            return Response(status=404)
        return Response(task)


@resource('/workflow/history/{uid}/tasks/{task_id}/data', versions=['v1'])
class ApiWorkflowHistoryTaskData:

    async def get(self, request, uid, task_id):
        try:
            task = await self.nyuki.storage.get_instance_task_data(task_id)
        except AutoReconnect:
            return Response(status=503)
        if not task:
            return Response(status=404)
        return Response(task)


@resource('/workflow/triggers', versions=['v1'])
class ApiWorkflowTriggers:

    async def get(self, request):
        """
        Return the list of all trigger forms
        """
        try:
            triggers = await self.nyuki.storage.triggers.get()
        except AutoReconnect:
            return Response(status=503)
        return Response(triggers)

    @content_type('multipart/form-data')
    async def put(self, request):
        """
        Upload a trigger form file
        """
        data = await request.post()
        try:
            form = data['form']
            tid = data['tid']
        except KeyError:
            return Response(status=400, body={
                'error': "'form' and 'tid' are mandatory parameters"
            })
        if not isinstance(form, FileField):
            return Response(status=400, body={
                'error': "'form' field must be a file content"
            })

        content = form.file.read().decode('utf-8')
        try:
            templates = await self.nyuki.storage.get_templates(template_id=tid)
            if not templates:
                return Response(status=404)
            trigger = await self.nyuki.storage.triggers.insert(tid, content)
        except AutoReconnect:
            return Response(status=503)
        return Response(trigger)


@resource('/workflow/triggers/{tid}', versions=['v1'])
class ApiWorkflowTrigger:

    async def get(self, request, tid):
        """
        Return a single trigger form
        """
        try:
            trigger = await self.nyuki.storage.triggers.get_one(tid)
        except AutoReconnect:
            return Response(status=503)
        if not trigger:
            return Response(status=404)
        return Response(trigger)

    async def delete(self, request, tid):
        """
        Delete a trigger form
        """
        try:
            trigger = await self.nyuki.storage.triggers.get_one(tid)
        except AutoReconnect:
            return Response(status=503)
        if not trigger:
            return Response(status=404)

        await self.nyuki.storage.triggers.delete(tid)
        return Response(trigger)



import logging
from copy import deepcopy

from motor.motor_asyncio import AsyncIOMotorClient
from pymongo.errors import ServerSelectionTimeoutError

from .triggers import TriggerCollection
from .data_processing import DataProcessingCollection
from .metadata import MetadataCollection
from .workflow_templates import WorkflowTemplatesCollection, TemplateState
from .task_templates import TaskTemplatesCollection
from .workflow_instances import WorkflowInstancesCollection
from .task_instances import TaskInstancesCollection


log = logging.getLogger(__name__)


class MongoStorage:

    def __init__(self):
        self._client = None
        self._db = None
        self._validate_on_start = False

        # Collections
        self._workflow_templates = None
        self._task_templates = None
        self._workflow_metadata = None
        self._workflow_instances = None
        self._task_instances = None
        self.regexes = None
        self.lookups = None
        self.triggers = None

    def configure(self, host, database, validate_on_start=True, **kwargs):
        log.info(
            "Setting up mongo storage with host '%s' and database '%s'",
            host, database,
        )
        self._client = AsyncIOMotorClient(host, **kwargs)
        self._db = self._client[database]
        self._validate_on_start = validate_on_start

        # Collections
        self._workflow_templates = WorkflowTemplatesCollection(self._db)
        self._task_templates = TaskTemplatesCollection(self._db)
        self._workflow_metadata = MetadataCollection(self._db)
        self._workflow_instances = WorkflowInstancesCollection(self._db)
        self._task_instances = TaskInstancesCollection(self._db)
        self.regexes = DataProcessingCollection(self._db, 'regexes')
        self.lookups = DataProcessingCollection(self._db, 'lookups')
        self.triggers = TriggerCollection(self._db)

    async def index(self):
        """
        Try to connect to mongo and index all the collections.
        """
        log.info('Trying to connect to Mongo...')
        while True:
            try:
                await self._workflow_templates.index()
                await self._task_templates.index()
                await self._workflow_metadata.index()
                await self._workflow_instances.index()
                await self._task_instances.index()
                await self.regexes.index()
                await self.lookups.index()
                await self.triggers.index()
            except ServerSelectionTimeoutError as exc:
                log.error('Could not connect to Mongo - %s', exc)
            else:
                log.info('Successfully connected to Mongo')
                break

        if self._validate_on_start is True:
            collections = await self._db.list_collection_names()
            log.info('Validating %s collections', len(collections))
            for collection in collections:
                await self._db.validate_collection(collection)
                log.info('Validated collection %s', collection)

    # Templates

    async def update_workflow_metadata(self, tid, metadata):
        """
        Update and return
        """
        return await self._workflow_metadata.update(tid, metadata)

    async def upsert_draft(self, template):
        """
        Update a template's draft and all its associated tasks.
        """
        metadata = await self._workflow_metadata.get_one(template['id'])
        if not metadata:
            metadata = {
                'workflow_template_id': template['id'],
                'title': template.pop('title'),
                'tags': template.pop('tags', []),
            }
            await self._workflow_metadata.insert(metadata)

        # Force set of values 'version' and 'draft'.
        template['version'] = await self._workflow_templates.get_last_version(
            template['id']
        ) + 1
        template['state'] = TemplateState.DRAFT.value

        # Split and insert tasks.
        tasks = template.pop('tasks')
        await self._task_templates.insert_many(deepcopy(tasks), template)

        # Insert template without tasks.
        await self._workflow_templates.insert_draft(template)
        template['tasks'] = tasks
        template.update({'title': metadata['title'], 'tags': metadata['tags']})
        return template

    async def publish_draft(self, template_id):
        """
        Publish a draft into an 'active' state, and archive the old active.
        """
        await self._workflow_templates.publish_draft(template_id)
        log.info('Draft for template %s published', template_id[:8])

    async def get_for_topic(self, topic):
        """
        Return all the templates listening on a particular topic.
        This does not append the metadata.
        """
        templates = await self._workflow_templates.get_for_topic(topic)
        if templates:
            log.info(
                'Fetched %s templates for event from "%s"',
                len(templates), topic,
            )
        for template in templates:
            template['tasks'] = await self._task_templates.get(
                template['id'], template['version']
            )
        return templates

    async def get_templates(self, template_id=None, full=False):
        """
        Return all active/draft templates
        Limited to a small set of fields if 'full' is False.
        TODO: Pagination.
        """
        templates = await self._workflow_templates.get(template_id, full)
        for template in templates:
            metadata = await self._workflow_metadata.get_one(template['id'])
            template.update(metadata)
            if full is True:
                template['tasks'] = await self._task_templates.get(
                    template['id'], template['version']
                )
        return templates

    async def get_template(self, tid, draft=False, version=None):
        """
        Return the active template.
        """
        template = await self._workflow_templates.get_one(
            tid,
            draft=draft,
            version=int(version) if version else None,
        )
        if not template:
            return

        metadata = await self._workflow_metadata.get_one(tid)
        template.update(metadata)
        template['tasks'] = await self._task_templates.get(
            template['id'], template['version']
        )
        return template

    async def delete_template(self, tid, draft=False):
        """
        Delete a whole template or only its draft.
        """
        await self._workflow_templates.delete(tid, draft)
        if draft is False:
            await self._task_templates.delete_many(tid)
            await self._workflow_metadata.delete(tid)
            await self.triggers.delete(tid)

    # Instances

    async def insert_instance(self, instance):
        """
        Insert a static workflow instance and all its tasks.
        """
        task_instances = []
        for task in instance['template'].pop('tasks'):
            task['workflow_instance_id'] = instance['id']
            task_instances.append(task)
        await self._task_instances.insert_many(task_instances)
        await self._workflow_instances.insert(instance)

    # History

    async def get_history(self, **kwargs):
        """
        Return paginated workflow history.
        """
        count, workflows = await self._workflow_instances.get(**kwargs)
        if kwargs.get('full') is True:
            for workflow in workflows:
                workflow['template']['tasks'] = await self._task_instances.get(
                    workflow['id'], True
                )
        return count, workflows

    async def get_instance(self, instance_id, full=False):
        workflow = await self._workflow_instances.get_one(instance_id, full)
        if not workflow:
            return
        workflow['template']['tasks'] = await self._task_instances.get(
            workflow['id'], full
        )
        return workflow

    async def get_instance_task(self, task_id, full=False):
        return await self._task_instances.get_one(task_id, full)

    async def get_instance_task_data(self, task_id):
        return await self._task_instances.get_data(task_id)



import os
from importlib import util


async def run_migrations(*args, **kwargs):
    folder = os.path.dirname(__file__)
    migrations = set()
    with os.scandir(folder) as sd:
        for entry in sd:
            if entry.name.startswith('__') or entry.name.endswith('__.py'):
                continue
            migrations.add(entry.name)

    for filename in migrations:
        name, _ = os.path.splitext(filename)
        path = os.path.join(folder, filename)
        # Load module spec.
        spec = util.spec_from_file_location(name, path)
        # Load module.
        module = spec.loader.load_module(spec.name)
        # Instantiate module migration and run.
        migration = module.Migration(*args, **kwargs)
        await migration.run()



import asyncio
import json
import logging
import re
from collections import namedtuple
from copy import copy
from uuid import uuid4

from hbmqtt.client import MQTTClient, ConnectException, ClientException
from hbmqtt.errors import NoDataException
from hbmqtt.mqtt.constants import QOS_0, QOS_1, QOS_2
from nyuki.services import Service
from nyuki.utils import serialize_object
from yarl import URL


log = logging.getLogger(__name__)

MQTTSubRegex = namedtuple('MQTTSubRegex', ['regex', 'callbacks'])


class MqttBus(Service):

    CONF_SCHEMA = {
        'type': 'object',
        'required': ['bus'],
        'properties': {
            'bus': {
                'type': 'object',
                'properties': {
                    'dsn': {'type': 'string', 'minLength': 1},
                    'cafile': {'type': 'string', 'minLength': 1},
                    'certfile': {'type': 'string', 'minLength': 1},
                    'keyfile': {'type': 'string', 'minLength': 1},
                    'keep_alive': {'type': 'integer', 'minimum': 1},
                    'ping_delay': {'type': 'integer', 'minimum': 1}
                },
                'additionalProperties': False
            }
        }
    }

    def __init__(self, nyuki, loop=None):
        self._nyuki = nyuki
        self._nyuki.register_schema(self.CONF_SCHEMA)
        self._loop = loop or asyncio.get_event_loop()
        self._dsn = None
        self.client = None
        self._subscriptions = {}
        self._regex_subscriptions = {}
        self._persisted = {}

        # Coroutines
        self.connect_future = None
        self.listen_future = None

    @property
    def topics(self):
        return list(self._subscriptions.keys())

    @property
    def name(self):
        return self._dsn.user

    def configure(self, dsn, cafile=None, certfile=None, keyfile=None,
                  keep_alive=60, ping_delay=5):
        self._dsn = URL(dsn)
        if self._dsn.scheme in ['mqtts', 'wss']:
            if not cafile or not certfile or not keyfile:
                raise ValueError(
                    "secured scheme requires 'cafile', 'certfile' and 'keyfile'"
                )

        self._cafile = cafile
        self.client = MQTTClient(
            client_id=f'{self.name}-{self._nyuki.id}',
            config={
                'auto_reconnect': False,
                'certfile': certfile,
                'keyfile': keyfile,
                'keep_alive': keep_alive,
                'ping_delay': ping_delay
            },
            loop=self._loop
        )

    async def start(self):
        def cancelled(future):
            try:
                future.result()
            except asyncio.CancelledError:
                log.debug('future cancelled: %s', future)
        self.connect_future = asyncio.ensure_future(self._run())
        self.connect_future.add_done_callback(cancelled)

    async def stop(self):
        # Clean client
        if self.client is not None:
            for task in self.client.client_tasks:
                log.debug('cancelling mqtt client tasks')
                task.cancel()
            if self.client._connected_state.is_set():
                log.debug('disconnecting mqtt client')
                await self.client.disconnect()
        # Clean tasks
        if self.connect_future:
            log.debug('cancelling _run coroutine')
            self.connect_future.cancel()
        if self.listen_future:
            log.debug('cancelling _listen coroutine')
            self.listen_future.cancel()
        log.info('MQTT service stopped')

    def _regex_topic(self, topic):
        """
        Transform the MQTT pattern into a regex object.
        """
        return re.compile(r'^{}$'.format(
            topic.replace('+', '[^\/]+').replace('#', '.+')
        ))

    async def subscribe(self, topic, callback):
        """
        Subscribe to a topic and setup the callback.
        This will pre-compile the regex used for this topic.
        """
        if not asyncio.iscoroutinefunction(callback):
            raise ValueError('event callback must be a coroutine')

        sub = False
        log.debug('MQTT subscription to %s -> %s', topic, callback.__name__)
        # Regexes are about topics like 'word/+/word' or 'word/#'
        is_regex = '+' in topic or topic.endswith('#')
        if is_regex is True:
            try:
                self._regex_subscriptions[topic].callbacks.add(callback)
            except KeyError:
                self._regex_subscriptions[topic] = MQTTSubRegex(
                    regex=self._regex_topic(topic),
                    callbacks={callback},
                )
                sub = True
        # Standard topics are a simple dict/set pair.
        else:
            try:
                self._subscriptions[topic].add(callback)
            except KeyError:
                self._subscriptions[topic] = {callback}
                sub = True

        # Send the subscription packet only if we were not subscribed yet
        if sub is True:
            await self.client.subscribe([(topic, QOS_1)])
            log.info('Subscribed to %s', topic)

        # Look for persisted message received prior to this subscription.
        if topic in self._persisted:
            for data in self._persisted[topic]:
                self._handle_message(topic, data)
            del self._persisted[topic]

    async def _unsub_regex(self, topic, callback):
        """
        Unsubscribe from a regex topic.
        """
        if topic not in self._regex_subscriptions:
            return
        if callback in self._regex_subscriptions[topic].callbacks:
            log.debug(
                'MQTT unsubscription from %s -> %s',
                topic, callback.__name__,
            )
            self._regex_subscriptions[topic].callbacks.remove(callback)
        if callback is None or not self._regex_subscriptions[topic].callbacks:
            del self._regex_subscriptions[topic]
            await self.client.unsubscribe([topic])
            log.info('Unsubscribed from %s', topic)

    async def _unsub(self, topic, callback):
        """
        Unsubscribe from a standard topic.
        """
        if topic not in self._subscriptions:
            return
        if callback in self._subscriptions[topic]:
            log.debug(
                'MQTT unsubscription from %s -> %s',
                topic, callback.__name__,
            )
            self._subscriptions[topic].remove(callback)
        if callback is None or not self._subscriptions[topic]:
            del self._subscriptions[topic]
            await self.client.unsubscribe([topic])
            log.info('Unsubscribed from %s', topic)

    async def unsubscribe(self, topic, callback=None):
        """
        Unsubscribe from a topic, remove callback if set.
        """
        if '+' in topic or topic.endswith('#'):
            await self._unsub_regex(topic, callback)
        else:
            await self._unsub(topic, callback)

    async def _resubscribe(self):
        """
        Resubscribe on reconnection.
        """
        # Resubscribe in case the MQTT broker restarted.
        subs = [topic for topic in self._subscriptions.keys()]
        subs.extend(self._regex_subscriptions.keys())
        for topic in subs:
            log.info('Resubscribing to %s', topic)
            await self.client.subscribe([(topic, QOS_1)])

    async def publish_qos_0(self, data, topic):
        return await self.publish(data, topic, qos=QOS_0)

    async def publish_qos_1(self, data, topic):
        return await self.publish(data, topic, qos=QOS_1)

    async def publish_qos_2(self, data, topic):
        return await self.publish(data, topic, qos=QOS_2)

    async def publish(self, data, topic=None, qos=QOS_0):
        """
        Publish in given topic or default one
        """
        if not topic:
            topic = self.name

        log.debug("Publishing event to '%s': %s", topic, data)
        data = json.dumps(data, default=serialize_object)

        if self.client._connected_state.is_set():
            try:
                await self.client.publish(topic, data.encode(), qos=qos)
            except Exception as exc:
                log.error('Error while publishing: %s', exc)
            else:
                log.debug('Event successfully sent to topic %s', topic)
        else:
            log.error('Failed to send event to topic %s', topic)

    async def _run(self):
        """
        Handle reconnection every 3 seconds
        """
        while True:
            log.info('Trying MQTT connection to %s', self._dsn)
            try:
                await self.client.connect(
                    str(self._dsn),
                    cleansession=False,
                    cafile=self._cafile,
                )
            except (ConnectException, NoDataException) as exc:
                log.error(exc)
                log.info('Waiting 3 seconds to reconnect')
                await asyncio.sleep(3.0)
                continue

            log.info('Connection made with MQTT')
            # Start listening.
            await self._resubscribe()
            self.listen_future = asyncio.ensure_future(self._listen())
            # Ensure the _persisted dict is emptied at some point.
            self._loop.call_later(60, self._clear_persisted)
            # Blocks until mqtt is disconnected.
            await self.client._handler.wait_disconnect()
            # Clean listen_future.
            self.listen_future.cancel()
            self.listen_future = None

    def _clear_persisted(self):
        self._persisted = {}

    async def _listen(self):
        """
        Listen to events after a successful connection.
        """
        while True:
            try:
                message = await self.client.deliver_message()
            except ClientException as exc:
                log.error(exc)
                break

            if message is None:
                log.info('listening loop ended')
                break

            self._handle_message(
                message.topic,
                json.loads(message.data.decode()),
            )

    def _handle_message(self, topic, data):
        handled = False
        # Iterate and call all regex topics callbacks
        for mqttregex in self._regex_subscriptions.values():
            if mqttregex.regex.match(topic):
                for callback in mqttregex.callbacks:
                    asyncio.ensure_future(callback(topic, data.copy()))
                handled = True

        # Iterate and call all single topic callbacks
        if topic in self._subscriptions:
            for callback in self._subscriptions[topic]:
                asyncio.ensure_future(callback(topic, data.copy()))
            handled = True

        # If the message was not linked to any known topic, keep it
        # in memory for a later subscription (should happen in an instant).
        # The dict is cleared 60 seconds after the connection.
        if handled is False:
            if topic in self._persisted:
                self._persisted[topic].append(data)
            else:
                self._persisted[topic] = [data]



import logging

from pymongo import ReturnDocument

from .utils.indexes import check_index_names


log = logging.getLogger(__name__)


class MetadataCollection:

    """
    {
        "workflow_template_id": <uuid4>,
        "title": <str>,
        "tags": [<str>]
    }
    """

    def __init__(self, db):
        self._metadata = db['workflow_metadata']

    async def index(self):
        await check_index_names(self._metadata, ['unique_workflow_template_id'])
        await self._metadata.create_index(
            'workflow_template_id',
            unique=True,
            name='unique_workflow_template_id',
        )

    async def get_one(self, tid):
        """
        Return metadata for one template.
        """
        return await self._metadata.find_one(
            {'workflow_template_id': tid},
            {'_id': 0, 'workflow_template_id': 0},
        )

    async def insert(self, metadata):
        """
        Insert new metadata for a template.
        """
        await self._metadata.insert_one(metadata)
        del metadata['_id']
        return metadata

    async def update(self, tid, metadata):
        """
        Update and return the updated metadata.
        """
        return await self._metadata.find_one_and_update(
            {'workflow_template_id': tid},
            {'$set': {
                key: value
                for key, value in metadata.items()
                if value is not None
            }},
            projection={'_id': 0},
            return_document=ReturnDocument.AFTER,
        )

    async def delete(self, tid):
        """
        Delete metadata for one template.
        """
        await self._metadata.delete_one({'workflow_template_id': tid})



import json
import asyncio
import logging
from enum import Enum
from aiohttp import ClientSession
from tukio.task import register
from tukio.task.holder import TaskHolder
from tukio.workflow import WorkflowExecState, Workflow

from .utils import runtime
from .utils.uri import URI


log = logging.getLogger(__name__)


class WorkflowStatus(Enum):

    PENDING = 'pending'
    RUNNING = 'running'
    TIMEOUT = 'timeout'
    DONE = 'done'


@register('trigger_workflow', 'execute')
class TriggerWorkflowTask(TaskHolder):

    __slots__ = (
        'template', 'blocking', 'task', '_engine', 'data',
        'status', 'triggered_id', 'async_future',
    )

    SCHEMA = {
        'type': 'object',
        'required': ['template'],
        'additionalProperties': False,
        'properties': {
            'template': {
                'type': 'object',
                'required': ['service', 'id'],
                'additionalProperties': False,
                'properties': {
                    'service': {'type': 'string', 'minLength': 1},
                    'id': {'type': 'string', 'minLength': 1},
                    'draft': {'type': 'boolean', 'default': False},
                },
            },
            'blocking': {'type': 'boolean', 'default': True},
        },
    }

    def __init__(self, config):
        super().__init__(config)
        self.template = self.config['template']
        self.blocking = self.config.get('blocking', True)
        self.task = None
        self._engine = 'http://{}/{}/api/v1/workflow'.format(
            runtime.config.get('http_host', 'localhost'),
            self.template['service'],
        )

        # Reporting
        self.status = WorkflowStatus.PENDING.value
        self.data = None
        self.triggered_id = None
        self.async_future = None

    def report(self):
        return {
            'exec_id': self.triggered_id,
            'status': self.status,
        }

    async def async_exec(self, topic, data):
        log.debug(
            "Received data for async trigger_workflow in '%s': %s",
            topic, data,
        )
        if not self.async_future.done():
            self.async_future.set_result(data)
        await runtime.bus.unsubscribe(topic)

    async def execute(self, event):
        """
        Entrypoint execution method.
        """
        self.data = event.data
        self.task = asyncio.Task.current_task()
        is_draft = self.template.get('draft', False)

        # Send the HTTP request
        log.info('Triggering template %s%s on service %s', self.template['id'],
                 ' (draft)' if is_draft else '', self.template['service'])

        # Setup headers (set requester and exec-track to avoid workflow loops)
        workflow = runtime.workflows[Workflow.current_workflow().uid]
        parent = workflow.exec.get('requester')
        track = list(workflow.exec.get('track', []))
        if parent:
            track.append(parent)

        headers = {
            'Content-Type': 'application/json',
            'Referer': URI.instance(workflow.instance),
            'X-Surycat-Exec-Track': ','.join(track)
        }

        # Handle blocking trigger_workflow using mqtt
        if self.blocking:
            topic = '{}/async/{}'.format(runtime.bus.name, self.uid[:8])
            headers['X-Surycat-Async-Topic'] = topic
            headers['X-Surycat-Async-Events'] = ','.join([
                WorkflowExecState.END.value,
                WorkflowExecState.ERROR.value,
            ])
            self.async_future = asyncio.Future()
            await runtime.bus.subscribe(topic, self.async_exec)

            def _unsub(f):
                asyncio.ensure_future(runtime.bus.unsubscribe(topic))
            self.task.add_done_callback(_unsub)

        async with ClientSession() as session:
            # Compute data to send to sub-workflows
            url = '{}/vars/{}{}'.format(
                self._engine,
                self.template['id'],
                '/draft' if is_draft else '',
            )
            async with session.get(url) as response:
                if response.status != 200:
                    raise RuntimeError("Can't load template info")
                wf_vars = await response.json()
            lightened_data = {
                key: self.data[key]
                for key in wf_vars
                if key in self.data
            }

            params = {
                'url': '{}/instances'.format(self._engine),
                'headers': headers,
                'data': json.dumps({
                    'id': self.template['id'],
                    'draft': is_draft,
                    'inputs': lightened_data,
                })
            }
            async with session.put(**params) as response:
                if response.status != 200:
                    msg = "Can't process workflow template {} on {}".format(
                        self.template, self.nyuki_api
                    )
                    if response.status % 400 < 100:
                        reason = await response.json()
                        msg = "{}, reason: {}".format(msg, reason['error'])
                    raise RuntimeError(msg)
                resp_body = await response.json()
                self.triggered_id = resp_body['id']

        wf_id = '@'.join([self.triggered_id[:8], self.template['service']])
        self.status = WorkflowStatus.RUNNING.value
        log.info('Successfully started %s', wf_id)
        self.task.dispatch_progress(self.report())

        # Block until task completed
        if self.blocking:
            log.info('Waiting for workflow %s to complete', wf_id)
            await self.async_future
            self.status = WorkflowStatus.DONE.value
            log.info('Workflow %s is done', wf_id)
            self.task.dispatch_progress({'status': self.status})

        return self.data

    async def _end_triggered_workflow(self):
        """
        Asynchronously cancel the triggered workflow.
        """
        wf_id = '@'.join([self.triggered_id[:8], self.template['service']])
        async with ClientSession() as session:
            url = '{}/instances/{}'.format(self._engine, self.triggered_id)
            async with session.delete(url) as response:
                if response.status != 200:
                    log.warning('Failed to cancel workflow %s', wf_id)
                else:
                    log.info('Workflow %s has been cancelled', wf_id)

    def teardown(self):
        """
        Called when this task is cancelled or timed out.
        """
        if self.task.timed_out is True:
            self.status = WorkflowStatus.TIMEOUT.value
        else:
            self.status = WorkflowStatus.DONE.value

        self.task.dispatch_progress({'status': self.status})
        if not self.triggered_id:
            log.debug('No workflow to cancel')
            return self.data

        asyncio.ensure_future(self._end_triggered_workflow())
        return self.data



import logging


log = logging.getLogger(__name__)


async def check_index_names(collection, names):
    # Add the default index on '_id'
    names.append('_id_')
    async for index in collection.list_indexes():
        if index['name'] not in names:
            log.warning(f"Bad index found in '{collection.name}', rebuilding")
            await collection.drop_indexes()
            break



import asyncio
import logging
from uuid import uuid4
from copy import deepcopy
from random import shuffle
from datetime import datetime
from tukio import Engine, TaskRegistry, get_broker, EXEC_TOPIC
from tukio.workflow import Workflow, WorkflowExecState
from tukio.task.factory import TaskExecState

from nyuki import Nyuki
from nyuki.utils import serialize_object, utcnow
from nyuki.workflow.db.storage import MongoStorage
from nyuki.workflow.db.migrations import run_migrations
from nyuki.workflow.db.task_instances import WS_FILTERS

from .api.factory import (
    ApiFactoryRegex, ApiFactoryRegexes, ApiFactoryLookup, ApiFactoryLookups,
    ApiFactoryLookupCSV
)
from .api.templates import (
    ApiTasks, ApiTemplates, ApiTemplate, ApiTemplateVersion, ApiTemplateDraft
)
from .api.instances import (
    ApiWorkflow, ApiWorkflows, ApiWorkflowsHistory, ApiWorkflowHistory,
    ApiWorkflowTriggers, ApiWorkflowTrigger, ApiWorkflowHistoryTask,
    ApiWorkflowHistoryTaskData, ApiTaskReporting, ApiTaskReportingContact,
    ApiTaskReportingContacts
)
from .api.vars import (
    ApiVars, ApiVarsVersion, ApiVarsDraft
)

from .tasks import *
from .tasks.utils import runtime, CONTACT_PROGRESS
from .tukio import WorkflowSelector


log = logging.getLogger(__name__)


class BadRequestError(Exception):
    pass


@serialize_object.register(Workflow)
def _serialize_workflow(wf):
    """
    Workflow serializer.
    """
    return wf.report()


def sanitize_workflow_exec(obj):
    """
    Replace any object value by 'internal data' string to store in Mongo.
    """
    types = [dict, list, tuple, str, int, float, bool, type(None), datetime]
    if type(obj) not in types:
        obj = 'Internal server data: {}'.format(type(obj))
    elif isinstance(obj, dict):
        for key, value in obj.items():
            obj[key] = sanitize_workflow_exec(value)
    elif isinstance(obj, list):
        for item in obj:
            item = sanitize_workflow_exec(item)
    return obj


class WorkflowInstance:

    """
    Holds a workflow pair of template/instance.
    Allows retrieving a workflow exec state at any moment.
    """

    __slots__ = ('_template', '_instance', '_exec')

    ALLOWED_EXEC_KEYS = ['requester', 'track']

    def __init__(self, template, instance, **kwargs):
        self._template = template
        self._instance = instance
        self._exec = {
            key: kwargs[key]
            for key in kwargs
            if key in self.ALLOWED_EXEC_KEYS
        }

    @property
    def template(self):
        return self._template

    @property
    def instance(self):
        return self._instance

    @property
    def exec(self):
        return self._exec

    def report(self, tasks=True, data=True):
        """
        Merge a workflow exec instance report and its template.
        """
        template = deepcopy(self._template)
        inst = self._instance.report()
        inst['exec'].update(self._exec)

        result = {
            **inst['exec'],
            'template': template,
        }

        if tasks is False:
            del result['template']['graph']
            del result['template']['tasks']
            return result

        tasks = {task['id']: {'template': task} for task in template['tasks']}
        for task_dict in inst['tasks']:
            if not task_dict.get('exec'):
                # Task was never started, create dummy exec dict.
                task_dict['exec'] = {
                    'id': str(uuid4()),
                    'start': None,
                    'end': None,
                    'state': 'not-started',
                    'inputs': None,
                    'outputs': None,
                    'reporting': None
                }

            # Filter out reporting/data if not necessary
            if data is False:
                del task_dict['exec']['reporting']
                del task_dict['exec']['inputs']
                # Leave the necessary task-end informations available
                if task_dict['exec']['outputs']:
                    task_dict['exec']['outputs'] = {
                        key: task_dict['exec']['outputs'][key]
                        for key in WS_FILTERS
                        if key in task_dict['exec']['outputs']
                    }

            # Add execution informations to each task.
            tasks[task_dict['id']].update(task_dict['exec'])

        result['template']['tasks'] = list(tasks.values())
        return result


class WorkflowNyuki(Nyuki):

    """
    Generic workflow nyuki allowing data storage and manipulation
    of tukio's workflows.
    https://github.com/surycat/tukio
    """

    CONF_SCHEMA = {
        'type': 'object',
        'required': ['mongo'],
        'properties': {
            'mongo': {
                'type': 'object',
                'required': ['host', 'database'],
                'properties': {
                    'host': {'type': 'string', 'minLength': 1},
                    'database': {'type': 'string', 'minLength': 1},
                    'validate_on_start': {'type': 'boolean', 'default': True},
                }
            },
            'topics': {
                'type': 'array',
                'items': {'type': 'string', 'minLength': 1}
            }
        }
    }
    HTTP_RESOURCES = Nyuki.HTTP_RESOURCES + [
        ApiTasks,                   # /v1/workflow/tasks
        ApiTemplates,               # /v1/workflow/templates
        ApiTemplate,                # /v1/workflow/templates/{uid}
        ApiTemplateDraft,           # /v1/workflow/templates/{uid}/draft
        ApiTemplateVersion,         # /v1/workflow/templates/{uid}/{version}
        ApiWorkflows,               # /v1/workflow/instances
        ApiWorkflow,                # /v1/workflow/instances/{uid}
        ApiTaskReporting,           # /v1/workflow/instances/{uid}/tasks/{task_id}/reporting
        ApiTaskReportingContacts,   # /v1/workflow/instances/{uid}/tasks/{task_id}/reporting/contacts
        ApiTaskReportingContact,    # /v1/workflow/instances/{uid}/tasks/{task_id}/reporting/contacts/{contact_id}
        ApiWorkflowsHistory,        # /v1/workflow/history
        ApiWorkflowHistory,         # /v1/workflow/history/{uid}
        ApiWorkflowHistoryTask,     # /v1/workflow/history/{uid}/tasks/{task_id}
        ApiWorkflowHistoryTaskData, # /v1/workflow/history/{uid}/tasks/{task_id}/data
        ApiFactoryRegexes,          # /v1/workflow/regexes
        ApiFactoryRegex,            # /v1/workflow/regexes/{uid}
        ApiFactoryLookups,          # /v1/workflow/lookups
        ApiFactoryLookup,           # /v1/workflow/lookups/{uid}
        ApiFactoryLookupCSV,        # /v1/workflow/lookups/{uid}/csv
        ApiWorkflowTriggers,        # /v1/workflow/triggers
        ApiWorkflowTrigger,         # /v1/workflow/triggers/{tid},
        ApiVars,                    # /v1/workflow/vars/{uid}
        ApiVarsVersion,             # /v1/workflow/vars/{uid}/{version}
        ApiVarsDraft,               # /v1/workflow/data/{uid}/draft
    ]

    DEFAULT_POLICY = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.migrate_config()
        self.register_schema(self.CONF_SCHEMA)
        self.schema = 1
        self.engine = None
        self.storage = MongoStorage()

        self.AVAILABLE_TASKS = {}
        for name, value in TaskRegistry.all().items():
            self.AVAILABLE_TASKS[name] = getattr(value[0], 'SCHEMA', {})

        # Stores workflow instances with their template data
        self.running_workflows = {}

        runtime.bus = self.bus
        runtime.config = self.config
        runtime.workflows = self.running_workflows

    @property
    def mongo_config(self):
        return self.config['mongo']

    @property
    def topics(self):
        return self.config.get('topics', [])

    async def setup(self):
        self.storage.configure(**self.mongo_config)
        # Blocks until connection to Mongo is done.
        await self.storage.index()
        await run_migrations(**self.mongo_config)
        selector = WorkflowSelector(self.storage)
        self.engine = Engine(selector=selector, loop=self.loop)
        for topic in self.topics:
            asyncio.ensure_future(self.bus.subscribe(
                topic, self.workflow_event
            ))
        # Enable workflow exec follow-up
        get_broker().register(self.report_workflow, topic=EXEC_TOPIC)

    async def reload(self):
        self.storage.configure(**self.mongo_config)

    async def teardown(self):
        if self.engine:
            await self.engine.stop()

    def new_workflow(self, template, instance, **kwargs):
        """
        Keep in memory a workflow template/instance pair.
        """
        wflow = WorkflowInstance(template, instance, **kwargs)
        self.running_workflows[instance.uid] = wflow
        return wflow

    async def report_workflow(self, event):
        """
        Send all worklfow updates to the clients.
        """
        source = event.source.as_dict()
        instance_id = source['workflow_exec_id']
        try:
            wflow = self.running_workflows[instance_id]
        except KeyError:
            log.debug('Outdated event to report: %s', event)
            return

        topic = 'workflow/exec/{}'.format(instance_id)
        payload = {
            'type': event.data['type'],
            'ts': utcnow(),
            'topic': topic,
            'service': self.bus.name,
        }

        # If is a workflow update, send the 'child' value
        if event.data['type'] in WorkflowExecState.values():
            requester = wflow.exec.get('requester')
            if requester:
                payload['requester'] = requester

        # A task information requires the corresponding template id
        # and a more precise topic.
        task_exec_id = source.get('task_exec_id')
        if task_exec_id:
            topic = '{}/tasks/{}'.format(topic, task_exec_id)
            payload['template'] = {'id': source['task_template_id']}

            # Append full data if this is a 'task-progress'
            if event.data['type'] == TaskExecState.PROGRESS.value:
                payload['data'] = event.data.get('content') or {}
                topic = '{}/reporting'.format(topic)

            # Custom event type for single contact updates
            elif event.data['type'] == CONTACT_PROGRESS:
                payload['data'] = event.data.get('content') or {}
                topic = '{}/reporting/contacts/{}'.format(
                    topic, payload['data']['uid'],
                )
                # Remove contact uid (available in topic)
                del payload['data']['uid']

            elif event.data['type'] in (
                TaskExecState.TIMEOUT.value,
                TaskExecState.END.value,
                TaskExecState.ERROR.value,
            ):
                if isinstance(event.data['content'], dict):
                    # Only send the task's important fields, if any
                    payload['data'] = {
                        key: event.data['content'][key]
                        for key in WS_FILTERS
                        if key in event.data['content']
                    }

            # Update topic for this event
            payload['topic'] = topic

        # Workflow begins, also send the full template.
        if event.data['type'] == WorkflowExecState.BEGIN.value:
            payload['template'] = dict(wflow.template)
        # Workflow ended, clear it from memory
        elif event.data['type'] in [
            WorkflowExecState.END.value,
            WorkflowExecState.ERROR.value
        ]:
            payload['data'] = event.data.get('content') or {}
            # Sanitize objects to store the finished workflow instance
            asyncio.ensure_future(self.storage.insert_instance(
                sanitize_workflow_exec(wflow.report())
            ))
            del self.running_workflows[instance_id]

        await self.bus.publish(payload, 'websocket/{}'.format(topic))

    async def workflow_event(self, topic, data):
        """
        New bus event received, trigger workflows if needed.
        """
        templates = {}
        # Retrieve full workflow templates
        # TODO: Better way to fetch the full template details
        # (this does may more requests than it should)
        wf_templates = await self.engine.selector.select(topic)
        for wftmpl in wf_templates:
            templates[wftmpl.uid] = await self.storage.get_template(
                wftmpl.uid, draft=False
            )
        # Trigger workflows
        instances = await self.engine.data_received(data, topic)
        for instance in instances:
            self.new_workflow(templates[instance.template.uid], instance)
